/*
**
** Sorting by Reversals
** --------------------
**
** Bergeron Greedy Algorithm.
**
**
** Date Started : Sunday, 30th May 2004.
** Date Finished: Tuesday, 29th June 2004.
**
*/



#include <stdio.h>
#include <malloc.h>
#include <math.h>		/* Necessary for ceil () function. */
#include <stdlib.h>		/* Necessary for rand () function. */
#include <time.h>		/* Necessary for time () function. */
#include "lists.h"


/*******************************************************************/

#define OK 1
#define SEED 200306

/*******************************************************************/

/* Dyadiko ena  : 00000001. */
const unsigned char ONE = ~(~(0^0)<<1);
/* Dyadiko mhden: 00000000. */
const unsigned char ZERO = 0^0;

/*******************************************************************/

struct bit_matrix{
	int			Dimension;
	int			words;
	unsigned char **	Matrix;
	unsigned char *		Parity;
	int *			Score;
};

typedef struct bit_matrix BIT_MATRIX;
typedef BIT_MATRIX * BIT_MATRIXPTR;



struct Permutation{
	int				Elements;
	int *			Array;
	BIT_MATRIXPTR	Bit_Matrix;
};

typedef struct Permutation PERMUTATION;
typedef PERMUTATION * PERMUTATIONPTR;



struct GreyEdge {
	unsigned int	start;
	unsigned int	finish;
};

typedef struct GreyEdge GREY_EDGE;
typedef GREY_EDGE * GREY_EDGEPTR;

/*******************************************************************/

PERMUTATIONPTR	Permutation_Initialize (const int);
int				Permutation_Destroy (PERMUTATIONPTR *);
int				Set_Identity_Permutation (PERMUTATIONPTR);
int				Reversal (PERMUTATIONPTR, const int, const int);
int				Permutation_Print (PERMUTATIONPTR);
int				Generate_Random_Permutation (PERMUTATIONPTR);
int				Sign (const int);
int				Initialize_Bit_Matrix (PERMUTATIONPTR);
int				Initialize_Bit_Matrix_Dimis (PERMUTATIONPTR);

/*******************************************************************/

/* Primitive Bit-Manipulation Functions. */
char create_char_from_binary (short int []);
void retrieve_binary_coding (const char, short int []);
char interchange_bit (const char, int);

/*******************************************************************/



int	main (void)
{
	PERMUTATIONPTR	permutation, identity, strange;
	int				temp, size;
	short int		pinakas_bit [sizeof (unsigned char) * 8];
	int				i, j, k;

	/*srand ( time (NULL) );*/
	srand (SEED);

	size = 13;
	permutation = Permutation_Initialize (size);
	identity = Permutation_Initialize (size);
	Set_Identity_Permutation (identity);
	Permutation_Print (identity);

	/* Generate hurdles. */
	printf ("\nGenerating hurdles...\n\n");
	/*
	** Generate sequence: 0 2 1 3.
	*/
	Reversal (identity, 1, 2);
	Reversal (identity, 2, 2);
	Reversal (identity, 1, 1);
	/*
	** This is tricky:
	** Generates sequence: 5 7 8 6 9
	** which is not a hurdle (7 and 8 are consecutive).
	*/
	Reversal (identity, 6, 8);
	Reversal (identity, 8, 8);
	Reversal (identity, 6, 7);
	/**/

	Permutation_Print (identity);

	printf ("\n");
	fflush (stdout);

	/*
	** Generate a rendom permutation.
	*/
	Generate_Random_Permutation (permutation);
	printf ("\nRandomly Generated Permutation:\n");
	Permutation_Print (permutation);
	
	printf ("\n");
	fflush (stdout);


	/*
	** Performance on cyclic hurdles.
	*/

	strange = Permutation_Initialize (6);
	strange->Array [1] = 3;
	strange->Array [2] = 1;
	strange->Array [3] = 6;
	strange->Array [4] = 5;
	strange->Array [5] = -2;
	strange->Array [6] = 4;

	Permutation_Print (strange);

	Initialize_Bit_Matrix_Dimis (strange);

	printf ("\n\nPrinting Bit-Matrix:\n\n");
	for (i = 0; i < strange->Bit_Matrix->Dimension; i++) {
		printf ("<%2d, %2d>: ", 2 * i, 2 * i + 1);
		for (j = 0; j < strange->Bit_Matrix->words; j++) {
			retrieve_binary_coding (strange->Bit_Matrix->Matrix [i][j], pinakas_bit);
			for (k = 0; k < 8; k++) {
				if (8 * j + k + 1 > strange->Bit_Matrix->Dimension)
					break;
				printf ("%d ", pinakas_bit [k]);
			}
		}
		printf ("\n");
	}
	printf ("\nParity vector:\n");
	printf ("          ");
	for (j = 0; j < strange->Bit_Matrix->words; j++) {
		retrieve_binary_coding (strange->Bit_Matrix->Parity [j], pinakas_bit);
		for (k = 0; k < 8; k++) {
			if (8 * j + k + 1 > strange->Bit_Matrix->Dimension)
				break;
			printf ("%d ", pinakas_bit [k]);
		}
	}
	printf ("\n");
	printf ("Score vector:\n");
	printf ("          ");
	for (j = 0; j < strange->Bit_Matrix->Dimension; j++)
		printf ("%d ", strange->Bit_Matrix->Score [j]);
	printf ("\n");


	printf ("\n");
	fflush (stdout);

	temp = Permutation_Destroy (&permutation);
	temp = Permutation_Destroy (&identity);
	temp = Permutation_Destroy (&strange);

/*
	printf ("\nsizeof (int)         : %d\n", sizeof (int));
	printf ("sizeof (unsigned int): %d\n", sizeof (unsigned int));
	printf ("ceil (%.6f)      : %.6f\n", 3.99, ceil (3.99));
*/

	return	0;
}



/*******************************************************************/

PERMUTATIONPTR	Permutation_Initialize (const int Elements)
{
	PERMUTATIONPTR	newPermutation;
	int				i, j;

	newPermutation = (PERMUTATIONPTR) malloc (sizeof (PERMUTATION));
	if (newPermutation == (PERMUTATIONPTR) NULL)
		return	(PERMUTATIONPTR) NULL;	/* Indicate error. */

	newPermutation->Elements = Elements;
	/* Reserve extra storage for 0 and EOP. */
	newPermutation->Array = (int *) malloc (sizeof (int) * (newPermutation->Elements + 2));
	if (newPermutation->Array == (int *) NULL) {
		free (newPermutation);

		return	(PERMUTATIONPTR) NULL;	/* Indicate error. */
	}

	/*
	** Try to reserve storage for entire bit-matrix structure,
	*/
	newPermutation->Bit_Matrix = (BIT_MATRIXPTR) malloc (sizeof (BIT_MATRIX));
	if (newPermutation->Bit_Matrix == (BIT_MATRIXPTR) NULL) {
		free (newPermutation->Array);
		free (newPermutation);

		return	(PERMUTATIONPTR) NULL;	/* Indicate error. */
	}

	/* Set Matrix Dimension. */
	newPermutation->Bit_Matrix->Dimension = Elements + 1;
	/* Find out how many words are required per line. */
	newPermutation->Bit_Matrix->words = (int) ceil ( ((double) Elements + 1) / (8.0 * ((double) sizeof (unsigned char))) );
	
	/* You need (Elements + 1) lines. */
	newPermutation->Bit_Matrix->Matrix = (unsigned char **) malloc (sizeof (unsigned char *) * (Elements + 1));
	/* Error checking. */
	if (newPermutation->Bit_Matrix->Matrix == (unsigned char **) NULL) {
		free (newPermutation->Bit_Matrix);
		free (newPermutation->Array);
		free (newPermutation);

		return	(PERMUTATIONPTR) NULL;	/* Indicate error. */
	}

	for (i = 0; i < (Elements + 1); i++) {
		/* Each line needs at most (8 * words) bits. */
		newPermutation->Bit_Matrix->Matrix [i] = (unsigned char *) malloc (sizeof (unsigned char) * newPermutation->Bit_Matrix->words);
		if (newPermutation->Bit_Matrix->Matrix [i] == (unsigned char *) NULL)
			break;
		for (j = 0; j < newPermutation->Bit_Matrix->words; j++)
			newPermutation->Bit_Matrix->Matrix [i][j] = (unsigned char) ZERO;
	}

	/* Error checking. */
	if ( (i < (Elements + 1)) && (newPermutation->Bit_Matrix->Matrix [i] == (unsigned char *) NULL) ) {
		while (i--)
			free (newPermutation->Bit_Matrix->Matrix [i]);
		free (newPermutation->Bit_Matrix->Matrix);
		free (newPermutation->Bit_Matrix);
		free (newPermutation->Array);
		free (newPermutation);

		return	(PERMUTATIONPTR) NULL;	/* Indicate error. */
	}

	/* One bit for each one of the vertices to indicate parity. */
	newPermutation->Bit_Matrix->Parity = (unsigned char *) malloc (sizeof (unsigned char) * newPermutation->Bit_Matrix->words);
	/* Error checking. */
	if (newPermutation->Bit_Matrix->Parity == (unsigned char *) NULL) {
		for (i = Elements; i >= 0; i--)
			free (newPermutation->Bit_Matrix->Matrix [i]);
		free (newPermutation->Bit_Matrix->Matrix);
		free (newPermutation->Bit_Matrix);
		free (newPermutation->Array);
		free (newPermutation);

		return	(PERMUTATIONPTR) NULL;	/* Indicate error. */
	}
	for (j = 0; j < newPermutation->Bit_Matrix->words; j++)
		newPermutation->Bit_Matrix->Parity [j] = (unsigned char) ZERO;

	/* Keep the scores of each vertex to an unsigned int. */
	newPermutation->Bit_Matrix->Score = (int *) malloc (sizeof (int) * newPermutation->Bit_Matrix->Dimension);
	/* Error checking. */
	if (newPermutation->Bit_Matrix->Score == (int *) NULL) {
		free (newPermutation->Bit_Matrix->Parity);
		for (i = Elements; i >= 0; i--)
			free (newPermutation->Bit_Matrix->Matrix [i]);
		free (newPermutation->Bit_Matrix->Matrix);
		free (newPermutation->Bit_Matrix);
		free (newPermutation->Array);
		free (newPermutation);

		return	(PERMUTATIONPTR) NULL;	/* Indicate error. */
	}


	/*
	** Necessary storage reserved.
	** Procceed to successful finish.
	*/


	/* Set limits. */
	newPermutation->Array [0] = (int) 0;
	newPermutation->Array [newPermutation->Elements + 1] = (int) (newPermutation->Elements + 1);

	return	newPermutation;	/* Everything is OK. */
}

/*******************************************************************/

int				Permutation_Destroy (PERMUTATIONPTR * permutation)
{
	int		i;

	if ((*permutation) != (PERMUTATIONPTR) NULL) {
		free ((*permutation)->Bit_Matrix->Score);
		free ((*permutation)->Bit_Matrix->Parity);
		for (i = (*permutation)->Bit_Matrix->Dimension - 1; i >= 0; i--)
			free ((*permutation)->Bit_Matrix->Matrix [i]);
		free ((*permutation)->Bit_Matrix->Matrix);
		free ((*permutation)->Bit_Matrix);
		free ((*permutation)->Array);
		free ((*permutation));
		*permutation = (PERMUTATIONPTR) NULL;
	}

	return	OK;		/* Memory is free! */
}

/*******************************************************************/

int				Permutation_Print (PERMUTATIONPTR permutation)
{
	int		i;

	printf ("<%+d ", permutation->Array [0]);
	for (i = 1; i < permutation->Elements + 1; i++)
		printf ("%+d ", permutation->Array [i]);
	printf ("%+d>\n", permutation->Array [i]);

	return	OK;
}

/*******************************************************************/

int				Reversal (PERMUTATIONPTR permutation, const int start, const int finish)
{
	int		i, temp;

	/* Catch exceptions. */
	if (start > finish)
		return -1;		/* Incorrect ordering. */
	if ((start <= 0) || (finish >= permutation->Elements + 1)) {
		if ((start <= 0) && (finish >= permutation->Elements + 1))
			return -2;	/* Both are out of bounds. */
		else if (start <= 0)
			return -3;	/* Starting position invalid. */
		else
			return -4;	/* Finishing position invalid. */
	}

	for (i = 0; i < (int) ceil ((double) (finish + 1 - start) / 2.0); i++) {
		temp = permutation->Array [start + i];
		permutation->Array [start + i] = -permutation->Array [finish - i];
		permutation->Array [finish - i] = -temp;
	}

	return	OK;
}

/*******************************************************************/

int				Set_Identity_Permutation (PERMUTATIONPTR permutation)
{
	int		i = permutation->Elements + 1;

	while (--i)
		permutation->Array [i] = i;

	return	OK;		/* Completed successfully. */
}

/*******************************************************************/

int				Generate_Random_Permutation (PERMUTATIONPTR permutation)
{
	LISTPTR			List;
	LIST_NODEPTR	node;
	int				i, temp;

	List = LIST__Initialize ();
	for (i = permutation->Elements; i >= 1; i--)
		LIST__Insert_node_at_front (List, i);

	for (i = 1; i <= permutation->Elements; i++) {
		temp = (rand () % List->NUMBER_OF_ELEMENTS) + 1;
		node = List->HEAD;
		while (--temp)
			node = node->next;
		if ((rand () % 2) == 0)
			permutation->Array [i] = node->key_code;
		else
			permutation->Array [i] = -node->key_code;

		LIST__Delete_node (List, node);
	}

	LIST__Destroy (List);

	/* Successful generation of random permutation. */

	return	OK;		
}

/*******************************************************************/

int				Sign (const int number)
{
	return (number >= 0) ? 1 : -1;
}

/*******************************************************************/

int				Initialize_Bit_Matrix (PERMUTATIONPTR permutation)
{
	register int		i, j, position;
	unsigned char *		a;
	short int			pinakas_bit [sizeof (char) * 8];
	register int		number, words, byte;

	words = permutation->Bit_Matrix->words;

	/* Reserve Memory. */
	a = (unsigned char *) malloc ( sizeof (unsigned char) * words);

	/* Scan permutation from Left to Right. */
	for (i = 0; i < words; i++)
		a [i] = ZERO;	/* Initialize to ZERO-bits (0000...0000) */

	a [0] = (unsigned char) interchange_bit (a [0], 1);	/* Initialize arc (0,1). */

	for (j = 1; j < permutation->Elements + 1; j++) {
		number = permutation->Array [j];

		if (number > 0) {
			/* Proccess arc (2*number - 2, 2*number - 1). */
			position = number;	/* Ari8mos bit apo aristera pros ta deksia. */

			byte = (position - 1) / (8 * sizeof (unsigned char));
			retrieve_binary_coding ( a [byte], pinakas_bit);
			/* Set a_j <- 1 if a_j == 0,  or  Set a_j <- 0 if a_j == 1. */
			a [byte] = (unsigned char) interchange_bit ( a [byte], ((position - 1) % 8) + 1);

			if (pinakas_bit [(position - 1) % 8]) {
				/* v_j <- a. */
				for (i = 0; i < permutation->Bit_Matrix->words; i++)
					permutation->Bit_Matrix->Matrix [position - 1][i] = a [i];
			}

			/* Proccess arc (2*number, 2*number + 1). */
			position = number + 1;		/* Ari8mos bit apo aristera pros ta deksia. */

			byte = (position - 1) / (8 * sizeof (unsigned char));
			retrieve_binary_coding ( a [byte], pinakas_bit);
			/* Set a_j <- 1 if a_j == 0,  or  Set a_j <- 0 if a_j == 1. */
			a [byte] = (unsigned char) interchange_bit ( a [byte], ((position - 1) % 8) + 1);

			if (pinakas_bit [(position - 1) % 8]) {
				/* v_j <- a. */
				for (i = 0; i < permutation->Bit_Matrix->words; i++)
					permutation->Bit_Matrix->Matrix [position - 1][i] = a [i];
			}
		}
		else {
			/* Proccess arc (2*number, 2*number + 1). */
			position = 1 - number;		/* Ari8mos bit apo aristera pros ta deksia. */

			byte = (position - 1) / (8 * sizeof (unsigned char));
			retrieve_binary_coding ( a [byte], pinakas_bit);
			/* Set a_j <- 1 if a_j == 0,  or  Set a_j <- 0 if a_j == 1. */
			a [byte] = (unsigned char) interchange_bit ( a [byte], ((position - 1) % 8) + 1);

			if (pinakas_bit [(position - 1) % 8]) {
				/* v_j <- a. */
				for (i = 0; i < permutation->Bit_Matrix->words; i++)
					permutation->Bit_Matrix->Matrix [position - 1][i] = a [i];
			}

			/* Proccess arc (2*number - 2, 2*number - 1). */
			position = -number;	/* Ari8mos bit apo aristera pros ta deksia. */

			byte = (position - 1) / (8 * sizeof (unsigned char));
			retrieve_binary_coding ( a [byte], pinakas_bit);
			/* Set a_j <- 1 if a_j == 0,  or  Set a_j <- 0 if a_j == 1. */
			a [byte] = (unsigned char) interchange_bit ( a [byte], ((position - 1) % 8) + 1);

			if (pinakas_bit [(position - 1) % 8]) {
				/* v_j <- a. */
				for (i = 0; i < permutation->Bit_Matrix->words; i++)
					permutation->Bit_Matrix->Matrix [position - 1][i] = a [i];
			}
		}
	}


	/* Scan permutation from Right to Left. */
	for (i = 0; i < words; i++)
		a [i] = ZERO;	/* Initialize to ZERO-bits (0000...0000) */

	/* Initialize arc (2*n, 2*n + 1). */
	a [words - 1] = (unsigned char) interchange_bit (a [words - 1], (permutation->Elements % 8) + 1);

	for (j = permutation->Elements; j > 0 ; j--) {
		number = permutation->Array [j];

		if (number > 0) {
			/* Proccess arc (2*number - 2, 2*number - 1). */
			position = number;	/* Ari8mos bit apo aristera pros ta deksia. */

			byte = (position - 1) / (8 * sizeof (unsigned char));
			retrieve_binary_coding ( a [byte], pinakas_bit);
			/* Set a_j <- 1 if a_j == 0,  or  Set a_j <- 0 if a_j == 1. */
			a [byte] = (unsigned char) interchange_bit ( a [byte], ((position - 1) % 8) + 1);

			if (pinakas_bit [(position - 1) % 8]) {
				/* v_j <- v_j XOR a. */
				for (i = 0; i < permutation->Bit_Matrix->words; i++)
					permutation->Bit_Matrix->Matrix [position - 1][i] ^= a [i];
			}

			/* Proccess arc (2*number, 2*number + 1). */
			position = number + 1;		/* Ari8mos bit apo aristera pros ta deksia. */

			byte = (position - 1) / (8 * sizeof (unsigned char));
			retrieve_binary_coding ( a [byte], pinakas_bit);
			/* Set a_j <- 1 if a_j == 0,  or  Set a_j <- 0 if a_j == 1. */
			a [byte] = (unsigned char) interchange_bit ( a [byte], ((position - 1) % 8) + 1);

			if (pinakas_bit [(position - 1) % 8]) {
				/* v_j <- v_j XOR a. */
				for (i = 0; i < permutation->Bit_Matrix->words; i++)
					permutation->Bit_Matrix->Matrix [position - 1][i] ^= a [i];
			}
		}
		else {
			/* Proccess arc (2*number, 2*number + 1). */
			position = 1 - number;		/* Ari8mos bit apo aristera pros ta deksia. */

			byte = (position - 1) / (8 * sizeof (unsigned char));
			retrieve_binary_coding ( a [byte], pinakas_bit);
			/* Set a_j <- 1 if a_j == 0,  or  Set a_j <- 0 if a_j == 1. */
			a [byte] = (unsigned char) interchange_bit ( a [byte], ((position - 1) % 8) + 1);

			if (pinakas_bit [(position - 1) % 8]) {
				/* v_j <- v_j XOR a. */
				for (i = 0; i < permutation->Bit_Matrix->words; i++)
					permutation->Bit_Matrix->Matrix [position - 1][i] ^= a [i];
			}

			/* Proccess arc (2*number - 2, 2*number - 1). */
			position = -number;	/* Ari8mos bit apo aristera pros ta deksia. */

			byte = (position - 1) / (8 * sizeof (unsigned char));
			retrieve_binary_coding ( a [byte], pinakas_bit);
			/* Set a_j <- 1 if a_j == 0,  or  Set a_j <- 0 if a_j == 1. */
			a [byte] = (unsigned char) interchange_bit ( a [byte], ((position - 1) % 8) + 1);

			if (pinakas_bit [(position - 1) % 8]) {
				/* v_j <- v_j XOR a. */
				for (i = 0; i < permutation->Bit_Matrix->words; i++)
					permutation->Bit_Matrix->Matrix [position - 1][i] ^= a [i];
			}
		}
	}


	/* Free-up memory reserved. */
	free (a);


	/* Successful Initialization of Bit-Matrix in permutation. */
	return	OK;
}

/*******************************************************************/

int				Initialize_Bit_Matrix_Dimis (PERMUTATIONPTR permutation)
{
	unsigned char **	r_matrix;
	unsigned char **	l_matrix;
	register int		i, j, k, words, lines, N, number, unoriented, oriented;
	unsigned int *		extended_permutation, * inv_ext, current_edge;
	GREY_EDGEPTR		greyEdges;
	register unsigned char 	mask, temp;
	short int 		pinakas_bit [8 * sizeof (char)];

	/* Set boundary values. */
	N = permutation->Elements;
	lines = permutation->Bit_Matrix->Dimension;
	words = permutation->Bit_Matrix->words;

	/* Reserve necessary memory - Initialize. */
	r_matrix = (unsigned char **) malloc (sizeof (unsigned char *) * lines);
	l_matrix = (unsigned char **) malloc (sizeof (unsigned char *) * lines);
	for (i = 0; i < lines; i++) {
		r_matrix [i] = (unsigned char *) malloc (sizeof (unsigned char) * words);
		l_matrix [i] = (unsigned char *) malloc (sizeof (unsigned char) * words);
		for (j = 0; j < words; j++) {
			/* Initialize to ZERO-bits. */
			r_matrix [i][j] = (unsigned char) ZERO;
			l_matrix [i][j] = (unsigned char) ZERO;
		}
	}

	/* Calculate extended and inverse extended permutation. */
	extended_permutation = (unsigned int *) malloc ( sizeof (unsigned int) * (2 * (N + 1)) );
	inv_ext = (unsigned int *) malloc ( sizeof (unsigned int) * (2 * (N + 1)) );
	extended_permutation [0] = 0;
	inv_ext [0] = 0;
	for (i = 1; i < N + 1; i++) {
		if (permutation->Array [i] > 0) {
			extended_permutation [2 * i - 1] = 2 * permutation->Array [i] - 1;
			extended_permutation [2 * i] = 2 * permutation->Array [i];
			inv_ext [2 * permutation->Array [i] - 1] = 2 * i - 1;
			inv_ext [2 * permutation->Array [i]] = 2 * i;
		}
		else {
			extended_permutation [2 * i - 1] = -2 * permutation->Array [i];
			extended_permutation [2 * i] = -2 * permutation->Array [i] - 1;
			inv_ext [-2 * permutation->Array [i]] = 2 * i - 1;
			inv_ext [-2 * permutation->Array [i] - 1] = 2 * i;
		}
	}
	extended_permutation [2 * N + 1] = 2 * N + 1;
	inv_ext [2 * N + 1] = 2 * N + 1;


	greyEdges = (GREY_EDGEPTR) malloc (sizeof (GREY_EDGE) * (N + 1));
	/* Calculate Grey Edges. */
	for (i = 0; i < (N + 1); i++) {
		if (inv_ext [2 * i] < inv_ext [2 * i + 1]) {
			greyEdges [i].start = inv_ext [2 * i];
			greyEdges [i].finish = inv_ext [2 * i + 1];
		}
		else {
			greyEdges [i].start = inv_ext [2 * i + 1];
			greyEdges [i].finish = inv_ext [2 * i];
		}
	}

	/* Calculate r_matrix. */
	for (i = 0; i < 2 * N + 2; i++) {
		number = extended_permutation [i];
		/* It belongs to greyEdge [floor (number / 2) ]. */
		if (greyEdges [number / 2].start == (unsigned int) i) {
			/* This is the beginning of the grey Edge. */
			for (j = i + 1; ((unsigned int) j) < greyEdges [number / 2].finish; j++) {
				current_edge = extended_permutation [j] / 2;
				if (greyEdges [current_edge].finish > greyEdges [number / 2].finish) {
					/* This is an intersection. */
					k = (current_edge % 8) + 1;
					r_matrix [number / 2][current_edge / 8] = (unsigned char) interchange_bit ( r_matrix [number / 2][current_edge / 8], k);
				}
			}
		}
	}

	/* Calculate l_matrix. */
	for (i = 2 * N + 1; i >= 0; i--) {
		number = extended_permutation [i];
		/* It belongs to greyEdge [floor (number / 2) ]. */
		if (greyEdges [number / 2].finish == (unsigned int) i) {
			/* This is the beginning of the grey Edge. */
			for (j = i - 1; ((unsigned int) j) > greyEdges [number / 2].start; j--) {
				current_edge = extended_permutation [j] / 2;
				if (greyEdges [current_edge].start < greyEdges [number / 2].start) {
					/* This is an intersection. */
					k = (current_edge % 8) + 1;
					l_matrix [number / 2][current_edge / 8] = (unsigned char) interchange_bit ( l_matrix [number / 2][current_edge / 8], k);
				}
			}
		}
	}

	/* Calculate Bit Matrix. */
	for (i = 0; i < lines; i++) {
		for (j = 0; j < words; j++) {
			permutation->Bit_Matrix->Matrix [i][j] = r_matrix [i][j] ^ l_matrix [i][j];
		}
	}

	/* Calculate Parity vector. */
	for (j = 0; j < words; j++) {
		permutation->Bit_Matrix->Parity [j] = (unsigned char) ZERO;
		for (i = 0; i < N + 1; i++) {
			permutation->Bit_Matrix->Parity [j] ^= permutation->Bit_Matrix->Matrix [i][j];
		}
	}


	/*
	** Calculate Score vector.
	** Remember that we keep only "variable" scores.
	** For further information please refer to the 1st (!)
	** edition of Bergeron's paper... :-)
	*/
	for (i = 0; i < N + 1; i++) {
		oriented = 0;
		unoriented = 0;
		for (j = 0; j < words; j++) {
			mask = permutation->Bit_Matrix->Parity [j]; 	/* Set mask. */
			temp = permutation->Bit_Matrix->Matrix [i][j] & mask;
			if ((i / 8 == j) && (temp & (ONE<<(7 - (i % 8))))) {
				/* vertex-bit is set to one. Set it to zero, so that you don't count this one. */
				temp ^= (ZERO << (7 - (i % 8)));
			}
			for (k = 0; k < 8; k++)
				if (temp & (ONE<<k))
					oriented++;
		}
		for (j = 0; j < words; j++) {
			mask = ~(permutation->Bit_Matrix->Parity [j]); 	/* Set mask. */
			temp = permutation->Bit_Matrix->Matrix [i][j] & mask;
			if ((i / 8 == j) && (temp & (ONE<<(7 - (i % 8)))))
				temp ^= (ZERO << (7 - (i % 8))); 	/* Do not count same vertex. (Reasons as above.) */
			for (k = 0; k < 8; k++)
				if (temp & (ONE<<k))
					unoriented++;
		}
		permutation->Bit_Matrix->Score [i] = (int) unoriented - oriented; 	/* Set "variable" score. */
	}




	/* Free-up extra memory. */
	for (i = lines - 1; i >= 0; i--) {
		free (r_matrix [i]);
		free (l_matrix [i]);
	}
	free (r_matrix);
	free (l_matrix);

	free (extended_permutation);
	free (inv_ext);
	free (greyEdges);


	/* Successful Initialization of Bit-Matrix in permutation. */
	return	OK;
}


/*******************************************************************/
/*******************************************************************/
/*******************************************************************/


char create_char_from_binary (
	short int pinakas_bit [sizeof (char) * 8]	/* Pinakas me th DYADIKH anaparastash */
												/* apo aristera pros ta deksia ayksanomenou */
												/* toy deikth 'i' tou pinaka. */
)

{
	unsigned char byte_created = 0^0;
	int i;

	for (i = 1; i <= 8; i++) {
		if (pinakas_bit [i - 1] == 1) {
			/* An theloume na exei 1 se ayth th 8esh */
			/* enallassoume to bit sth sygkekrimenh 8esh. */
			byte_created = interchange_bit (byte_created, i);
		}
	}

	return (char) byte_created;
}

/*******************************************************************/

void retrieve_binary_coding (
	const char	byte_to_be_decoded,			/* Byte tou opoiou 8a broume thn dyadikh anaparastash */
	short int	pinakas_bit [sizeof (char) * 8] /* Pinakas mege8ous sizeof (char) pou periexei thn    */
											/* dyadikh anaparastash toy byte (prwto orisma)       */
)

{
	unsigned char temp_byte = byte_to_be_decoded;	/* Antigrafh toy orismatos wste to orisma na */
													/* mhn metablh8ei me tis diafores olis8hseis */
	unsigned char mask = 1 << 7;					/* H maska h opoia 8a xrhsimopoih8ei. */
	int i;	/* Metrhths - Apari8mhths twn bit. */

	for (i = 1; i <= (sizeof (char) * 8); i++) {
		if (temp_byte & mask) {
			pinakas_bit [i - 1] = 1;
		}
		else {
			pinakas_bit [i - 1] = 0;
		}

		/* Olis8hse sto epomeno bit */
		temp_byte <<= 1;
	}
}

/*******************************************************************/

char interchange_bit (
	const char	byte_pros_allagh,			/* Byte tou opoiou 1 BIT 8a allaksoume */
	int			ari8mos_bit_pros_allagh		/* THESH BIT mesa sto BYTE (Aristera->Deksia) */
)

{
	unsigned char temp_byte = byte_pros_allagh;	/* Antigrafh tou xarakthra eisodou. */
	unsigned char mask = ONE << (8 - ari8mos_bit_pros_allagh);	/* Maska */

	return (char) (temp_byte^mask);
}

/*******************************************************************/
