#include "lists.h"

/* Necessary Initializations */
LISTPTR LIST__Initialize ()
{
	LISTPTR list = (LISTPTR) malloc (sizeof (LIST));

	/* NIL's Initialization */
	list->NIL = (LIST_NODEPTR) malloc (sizeof (LIST_NODE));
	list->NIL->key_code = NULL_KEY;
	list->NIL->next = list->NIL;
	list->NIL->previous = list->NIL;

	/* HEAD's Initialization */
	list->HEAD = list->NIL;

	/* Nothing is yet inserted */
	list->NUMBER_OF_ELEMENTS = 0;

	return list;	// Return new list.
}


/* List's Search Operation */
LIST_NODEPTR LIST__Search (LISTPTR list, KEY_TYPE key)
{
	LIST_NODEPTR current = list->HEAD;

	while ((current != list->NIL) && (current->key_code != key))
		current = current->next;	// Procceed to next element in list.

	return current;
}


/* Insert node at front - Do not allow duplicate insertions according to key */
int LIST__Insert_node_at_front (LISTPTR list, KEY_TYPE key)
{
	LIST_NODEPTR new_node = (LIST_NODEPTR) malloc (sizeof (LIST_NODE));
	LIST_NODEPTR search_node;

	search_node = LIST__Search (list, key);

	if (search_node != list->NIL) {
		free (new_node);	// Deallocate memory.

		if (search_node->key_code == key)
			return 0;		// Indicate key was previously inserted in list.
		else
			return -1;		// Indicate generic error.
	}

	/* Node's Initializations */
	new_node->key_code = key;

	/* List's Modifications */
	new_node->next = list->HEAD;
	
	if (list->HEAD != list->NIL)
		(list->HEAD)->previous = new_node;

	list->HEAD = new_node;
	new_node->previous = list->NIL;

	list->NUMBER_OF_ELEMENTS++;		// Increment counter;

	return 1;	// Indicate successful insertion.
}


/* Search List for a node with KEY == key and if found, delete it. */
int LIST__Delete_node_according_to_key (LISTPTR list, KEY_TYPE key)
{
	LIST_NODEPTR search_node;

	if (list->NUMBER_OF_ELEMENTS == 0)
		return -1;		// Indicate list is empty.

	search_node = LIST__Search (list, key);

	if (search_node == list->NIL)
		return 0;		// Node with corresponding key was not found.

	if (search_node->previous != list->NIL)
		(search_node->previous)->next = search_node->next;
	else
		list->HEAD = search_node->next;

	if (search_node->next != list->NIL)
		(search_node->next)->previous = search_node->previous;

	free (search_node);		// Deallocate memory.

	list->NUMBER_OF_ELEMENTS--;		// Decrement counter.

	return 1;	// Indicate successful deletion.
}

/* Delete a specific node in list. */
int LIST__Delete_node (LISTPTR list, LIST_NODEPTR node)
{
	if (list->NUMBER_OF_ELEMENTS == 0)
		return -1;		// Indicate list is empty.

	if (node == list->NIL)
		return 0;		// Incorrect node.

	if (node->previous != list->NIL)
		(node->previous)->next = node->next;
	else
		list->HEAD = node->next;

	if (node->next != list->NIL)
		(node->next)->previous = node->previous;

	free (node);		// Deallocate memory.

	list->NUMBER_OF_ELEMENTS--;		// Decrement counter.

	return 1;	// Indicate successful deletion.
}

/* Destroy List */
int LIST__Destroy (LISTPTR list)
{
	LIST_NODEPTR current;
	int key;

	if (list->NUMBER_OF_ELEMENTS > 0) {
		current = list->HEAD;
		key = current->key_code;
		while (list->NUMBER_OF_ELEMENTS > 0) {
			current = list->HEAD;
			key = current->key_code;
			LIST__Delete_node_according_to_key (list, key);
		}
	}

	/* Deallocate remaining memory. */
	free (list->NIL);
	free (list);

	return 1;	// Indicate successful destruction.
}


/* Swap Nodes in List */
int LIST__Swap_nodes (LISTPTR List, LIST_NODEPTR node1, LIST_NODEPTR node2)
{
	KEY_TYPE temp_key;

	if (node1 == List->NIL || node2 == List->NIL)
		return 0;	// Indicate no swaping occurred.

	temp_key = node2->key_code;

	/* Change contents in nodes */
	node2->key_code = node1->key_code;
	node1->key_code = temp_key;

	return 1;		// Indicate successful swaping.
}
