#include "rb-trees.h"

/*****************************************************************************/

int REALLOCATE__DATA_KEY (TREE_NODEPTR source, TREE_NODEPTR target)
{
	/* Necessary function for RED_BLACK__Delete function. */
	target->key = source->key;

	//target->data->number = source->data->number;

	return 1;
}

/*****************************************************************************/

TREEPTR TREE__Create ()
{
	TREEPTR Tree;

	Tree = (TREEPTR) malloc (sizeof (TREE));
	Tree->NIL = (TREE_NODEPTR) malloc (sizeof (TREE_NODE));
	//Tree->NIL->data = (DATA_TYPEPTR) malloc (sizeof (DATA_TYPE));
	Tree->ROOT = Tree->NIL;

	Tree->NIL->color = BLACK;
	Tree->NIL->parent = Tree->NIL;
	Tree->NIL->key = NULL_KEY;
	Tree->ROOT->parent = Tree->NIL;
	Tree->ROOT->left_child = Tree->NIL;
	Tree->ROOT->right_child = Tree->NIL;
	Tree->Elements = 0;

	return Tree;
}

TREE_NODEPTR TREE_NODE__Create (TREEPTR Tree)
{
	TREE_NODEPTR NewNode;

	NewNode = (TREE_NODEPTR) malloc (sizeof (TREE_NODE));
	//NewNode->data = (DATA_TYPEPTR) malloc (sizeof (DATA_TYPE));
	NewNode->left_child = Tree->NIL;
	NewNode->right_child = Tree->NIL;
	NewNode->parent = Tree->NIL;

	return NewNode;
}

int TREE_NODE__Destroy (TREE_NODEPTR OldNode)
{
	//free (OldNode->data);
	free (OldNode);

	return 1;	// Indicate successful deletion.
}

TREE_NODEPTR TREE__Search (TREEPTR Tree, KEY_TYPE k)
{
	TREE_NODEPTR x;

	x = Tree->ROOT;
	while ((x != Tree->NIL) && (x->key != k)) {
		if (k < x->key)
			x = x->left_child;
		else
			x = x->right_child;
	}

	return x;
}

TREE_NODEPTR TREE__Minimum (TREEPTR Tree, TREE_NODEPTR x)
{
	while (x->left_child != Tree->NIL)
		x = x->left_child;

	return x;
}

TREE_NODEPTR TREE__Maximum (TREEPTR Tree, TREE_NODEPTR x)
{
	while (x->right_child != Tree->NIL)
		x = x->right_child;

	return x;
}

TREE_NODEPTR TREE__Successor (TREEPTR Tree, TREE_NODEPTR x)
{
	TREE_NODEPTR y = NULL;

	if (x->right_child != Tree->NIL) {
		return TREE__Minimum (Tree, x->right_child);
	}

	y = x->parent;
	while ((y != Tree->NIL) && (x == y->right_child)) {
		x = y;
		y = y->parent;
	}

	return y;
}

TREE_NODEPTR TREE__Predecessor (TREEPTR Tree, TREE_NODEPTR x)
{
	TREE_NODEPTR y = NULL;

	if (x->left_child != Tree->NIL) {
		return TREE__Maximum (Tree, x->left_child);
	}

	y = x->parent;
	while ((y != Tree->NIL) && (x == y->left_child)) {
		x = y;
		y = y->parent;
	}

	return y;
}

/*****************************************************************************/

int TREE__Insert (TREEPTR Tree, TREE_NODEPTR new_node)
{
	int found = 0;
	TREE_NODEPTR x, y;

	y = Tree->NIL;
	x = Tree->ROOT;

	while (x != Tree->NIL) {
		y = x;
		if (new_node->key < x->key)
			x = x->left_child;
		else {
			if (new_node->key > x->key)
				x = x->right_child;
			else {	/* new_node->key == x->key */
				found = 1;
				break;
			}
		}
	}

	if (found)
		return 0;		/* No Insertion took place. */

	new_node->parent = y;
	if (y == Tree->NIL) {
		Tree->ROOT = new_node;
	}
	else {
		if (new_node->key < y->key) {
			y->left_child = new_node;
		}
		else {
			if (new_node->key > y->key) {
				y->right_child = new_node;
			}
			else {
				/* In this case: new_node->key == y->key */
				return 0;	/* No Insertion took place. */
			}
		}
	}

	return 1;	/* Succesful Insertion. */
}

/*****************************************************************************/

int LEFT__Rotate (TREEPTR Tree, TREE_NODEPTR x)
{
	/* 
	 * We assume that:
	 * x->right_child != Tree->NIL
	 */
	TREE_NODEPTR y;

	y = x->right_child;
	x->right_child = y->left_child;		/* Turn y's left subtree into x's right subtree. */
	if (y->left_child != Tree->NIL)
		(y->left_child)->parent = x;
	y->parent = x->parent;
	if (x->parent == Tree->NIL) {
		Tree->ROOT = y;
	}
	else {
		if (x == (x->parent)->left_child)
			(x->parent)->left_child = y;
		else
			(x->parent)->right_child = y;
	}
	y->left_child = x;		/* Put x on y's left. */
	x->parent = y;

	return 1;
}

int RIGHT__Rotate (TREEPTR Tree, TREE_NODEPTR x)
{
	/* 
	 * We assume that:
	 * x->left_child != Tree->NIL
	 */
	TREE_NODEPTR y;

	y = x->left_child;
	x->left_child = y->right_child;		/* Turn y's right subtree into x's left subtree. */
	if (y->right_child != Tree->NIL)
		(y->right_child)->parent = x;
	y->parent = x->parent;
	if (x->parent == Tree->NIL) {
		Tree->ROOT = y;
	}
	else {
		if (x == (x->parent)->left_child)
			(x->parent)->left_child = y;
		else
			(x->parent)->right_child = y;
	}
	y->right_child = x;		/* Put x on y's right. */
	x->parent = y;

	return 1;
}


int RED_BLACK__Insert (TREEPTR Tree, TREE_NODEPTR x)
{
	int binary_insertion;
	TREE_NODEPTR y = NULL;

	binary_insertion = TREE__Insert (Tree, x);
	if (binary_insertion == 0) {
		return 0;	/* Indicate no insertion took place. */
	}
	x->color = RED;
	while ((x != Tree->ROOT) && ((x->parent)->color == RED)) {
		if (x->parent == ((x->parent)->parent)->left_child) {
			y = ((x->parent)->parent)->right_child;
			if (y->color == RED) {
				(x->parent)->color = BLACK;
				y->color = BLACK;
				((x->parent)->parent)->color = RED;
				x = (x->parent)->parent;
			}
			else {
				if (x == (x->parent)->right_child) {
					x = x->parent;
					LEFT__Rotate (Tree, x);
				}
				(x->parent)->color = BLACK;
				((x->parent)->parent)->color = RED;
				RIGHT__Rotate (Tree, (x->parent)->parent);
			}
		}
		else {
			y = ((x->parent)->parent)->left_child;
			if (y->color == RED) {
				(x->parent)->color = BLACK;
				y->color = BLACK;
				((x->parent)->parent)->color = RED;
				x = (x->parent)->parent;
			}
			else {
				if (x == (x->parent)->left_child) {
					x = x->parent;
					RIGHT__Rotate (Tree, x);
				}
				(x->parent)->color = BLACK;
				((x->parent)->parent)->color = RED;
				LEFT__Rotate (Tree, (x->parent)->parent);
			}
		}
	}

	Tree->ROOT->color = BLACK;
	Tree->Elements++;

	return 1;	/* Indicate successful insertion. */
}

TREE_NODEPTR RED_BLACK__Delete (TREEPTR Tree, TREE_NODEPTR z)
{
	TREE_NODEPTR x, y;

	if ((z->left_child == Tree->NIL) || (z->right_child == Tree->NIL))
		y = z;
	else
		y = TREE__Successor (Tree, z);

	if (y->left_child != Tree->NIL)
		x = y->left_child;
	else
		x = y->right_child;

	x->parent = y->parent;

	if (y->parent == Tree->NIL) {
		Tree->ROOT = x;
	}
	else {
		if (y == (y->parent)->left_child) {
			(y->parent)->left_child = x;
		}
		else {
			(y->parent)->right_child = x;
		}
	}

	if (y != z)
		REALLOCATE__DATA_KEY (y, z);

	if (y->color == BLACK)
		RED_BLACK__Delete_FIXUP (Tree, x);

	Tree->Elements--;

	return y;	/* Return the address of an empty node. */
}

int RED_BLACK__Delete_FIXUP (TREEPTR Tree, TREE_NODEPTR x)
{
	TREE_NODEPTR w = Tree->NIL;

	while ((x != Tree->ROOT) && (x->color == BLACK)) {
		if (x == (x->parent)->left_child) {
			w = (x->parent)->right_child;
			if (w->color == RED) {
				w->color = BLACK;
				(x->parent)->color = RED;
				LEFT__Rotate (Tree, x->parent);
				w = (x->parent)->right_child;
			}
			if (((w->left_child)->color == BLACK) && ((w->right_child)->color == BLACK)) {
				w->color = RED;
				x = x->parent;
			}
			else {
				if ((w->right_child)->color == BLACK) {
					(w->left_child)->color = BLACK;
					w->color = RED;
					RIGHT__Rotate (Tree, w);
					w = (x->parent)->right_child;
				}
				w->color = (x->parent)->color;
				(x->parent)->color = BLACK;
				(w->right_child)->color = BLACK;
				LEFT__Rotate (Tree, x->parent);
				x = Tree->ROOT;
			}
		}
		else {
			w = (x->parent)->left_child;
			if (w->color == RED) {
				w->color = BLACK;
				(x->parent)->color = RED;
				RIGHT__Rotate (Tree, x->parent);
				w = (x->parent)->left_child;
			}
			if (((w->right_child)->color == BLACK) && ((w->left_child)->color == BLACK)) {
				w->color = RED;
				x = x->parent;
			}
			else {
				if ((w->left_child)->color == BLACK) {
					(w->right_child)->color = BLACK;
					w->color = RED;
					LEFT__Rotate (Tree, w);
					w = (x->parent)->left_child;
				}
				w->color = (x->parent)->color;
				(x->parent)->color = BLACK;
				(w->left_child)->color = BLACK;
				RIGHT__Rotate (Tree, x->parent);
				x = Tree->ROOT;
			}
		}
	}

	x->color = BLACK;

	return 1;	/* Proceed to succesful FIXUP */
}

/*********************************************************/
